import typing as t
from dataclasses import dataclass

from flow_prompt.ai_models.ai_model import AIModel


@dataclass
class AttemptToCall:
    ai_model: AIModel
    weight: int = 1  # from 1 to 100, the higher weight the more often it will be called
    # if you wish to limit functions that can be used, or to turn off calling openai functions for this attempt:
    # [] - if empty list of functions, functions are not supported for that call
    # None - if None, no limitations on functions
    # ['function1', 'function2'] - if list of functions, only those functions will be called
    functions: t.List[str] = None
    attempt_number: int = 1

    def __post_init__(self):
        self.id = (
            f"{self.ai_model.name}"
            f"-n{self.attempt_number}-"
            f"{self.ai_model.provider.value}"
        )

    def __str__(self) -> str:
        return self.id

    def params(self) -> t.Dict[str, t.Any]:
        self.ai_model.get_params()

    def get_functions(self) -> t.List[str]:
        # empty list - functions are not supported
        if not self.ai_model.support_functions:
            return []
        # None - no limitations on functions
        if self.functions is None:
            return None
        return self.functions

    def model_max_tokens(self) -> int:
        return self.ai_model.max_tokens

    def tiktoken_encoding(self) -> str:
        return self.ai_model.tiktoken_encoding



from flow_prompt.exceptions import FlowPromptException, RetryableCustomException


class OpenAIChunkedEncodingError(RetryableCustomException):
    pass


class OpenAITimeoutException(RetryableCustomException):
    pass


class OpenAIResponseWasFilteredError(RetryableCustomException):
    pass


class OpenAIAuthenticationError(RetryableCustomException):
    pass


class OpenAIInternalException(RetryableCustomException):
    pass


class OpenAiRateLimitException(RetryableCustomException):
    pass


class OpenAiPermissionDeniedError(RetryableCustomException):
    pass


class OpenAIUnknownException(RetryableCustomException):
    pass


### Non-retryable exceptions ###
class OpenAIInvalidRequestError(FlowPromptException):
    pass


class OpenAIBadRequestException(FlowPromptException):
    pass



import logging
import typing as t
from collections import defaultdict
from dataclasses import dataclass, field

import tiktoken

from flow_prompt import settings
from flow_prompt.exceptions import NotEnoughBudgetException
from flow_prompt.prompt.base_prompt import BasePrompt
from flow_prompt.prompt.chat import ChatMessage, ChatsEntity

logger = logging.getLogger(__name__)


@dataclass
class State:
    """
    State of the prompt. left_budget is the budget left for the rest of the prompt.
    fully_fitted_pipitas is the set of labels of chats that were fully fitted in the prompt.
    Pipita references to a small part of pipe, formed with a Spanish ending 'ita' which means a smaller version.
    """

    left_budget: int = 0
    fully_fitted_pipitas: t.Set[str] = field(default_factory=set)
    references: t.Dict[str, t.List[str]] = field(
        default_factory=lambda: defaultdict(list)
    )


@dataclass
class CallingMessages:
    messages: t.List[ChatMessage]
    prompt_budget: int = 0
    left_budget: int = 0
    references: t.Dict[str, t.List[str]] = None
    max_sample_budget: int = 0

    @property
    def calling_messages(self) -> t.List[t.Dict[str, str]]:
        return [m.to_dict() for m in self.messages if not m.is_empty()]

    def get_messages(self) -> t.List[t.Dict[str, str]]:
        result = []
        for m in self.messages:
            if m.is_empty():
                continue
            result.append(m.to_dict())
        return result

    def __str__(self) -> str:
        return "\n".join([str(m.to_dict()) for m in self.messages])


@dataclass(kw_only=True)
class UserPrompt(BasePrompt):
    model_max_tokens: int
    tiktoken_encoding: str
    min_sample_tokens: int
    reserved_tokens_budget_for_sampling: int = None
    safe_gap_tokens: int = settings.SAFE_GAP_TOKENS

    def __post_init__(self):
        self.encoding = tiktoken.get_encoding(self.tiktoken_encoding)

    def resolve(self, context: t.Dict) -> CallingMessages:
        pipe = {}
        prompt_budget = 0
        ordered_pipe = dict((value, i) for i, value in enumerate(self.pipe))
        state = State()
        state.left_budget = self.left_budget
        for priority in sorted(self.priorities.keys()):
            for chat_value in self.priorities[priority]:
                r = [
                    p in state.fully_fitted_pipitas
                    for p in (chat_value.add_if_fitted_labels or [])
                ]
                if not all(r):
                    continue

                if chat_value.presentation:
                    state.left_budget -= len(
                        self.encoding.encode(chat_value.presentation)
                    )
                if chat_value.last_words:
                    state.left_budget -= len(
                        self.encoding.encode(chat_value.last_words)
                    )

                values = chat_value.get_values(context)
                logger.debug(f"Got values for {chat_value}: {values}")
                if not values:
                    continue
                if chat_value.in_one_message:
                    messages_budget, messages = self.add_values_in_one_message(
                        values, chat_value, state
                    )
                elif chat_value.while_fits:
                    messages_budget, messages = self.add_values_while_fits(
                        values,
                        chat_value,
                        state,
                    )
                else:
                    messages_budget, messages = self.add_values(values, state)
                    if chat_value.label:
                        state.fully_fitted_pipitas.add(chat_value.label)

                if not messages:
                    logger.debug(f"messages is empty for {chat_value}")
                    continue
                if not self.is_enough_budget(state, messages_budget):
                    logger.debug(f"not enough budget for {chat_value}")
                    if chat_value.required:
                        raise NotEnoughBudgetException("Not enough budget")
                    continue
                logger.debug(f"adding {len(messages)} messages for {chat_value}")
                state.left_budget -= messages_budget
                prompt_budget += messages_budget
                if chat_value.presentation:
                    messages[0].content = chat_value.presentation + messages[0].content
                if chat_value.last_words:
                    messages[-1].content += chat_value.last_words
                pipe[chat_value._uuid] = messages
                continue

        final_pipe_with_order = [
            pipe.get(chat_id, [])
            for chat_id, _ in sorted(ordered_pipe.items(), key=lambda x: x[1])
        ]
        # skip empty values
        flat_list: t.List[ChatMessage] = [
            item for sublist in final_pipe_with_order for item in sublist if item
        ]
        max_sample_budget = left_budget = state.left_budget + self.min_sample_tokens
        if self.reserved_tokens_budget_for_sampling:
            max_sample_budget = min(
                self.reserved_tokens_budget_for_sampling, left_budget
            )
        return CallingMessages(
            references=state.references,
            messages=flat_list,
            prompt_budget=prompt_budget,
            left_budget=left_budget,
            max_sample_budget=max_sample_budget,
        )

    def add_values_while_fits(
        self,
        values: list[ChatMessage],
        chat_value: ChatsEntity,
        state: State,
    ):
        add_in_reverse_order = chat_value.add_in_reverse_order
        if add_in_reverse_order:
            values = values[::-1]
        values_to_add = []
        messages_budget = 0
        is_fully_fitted = True
        if not values:
            logger.info(
                f"[{self.task_name}]: values to add is empty {chat_value.prompt}"
            )
        for i, value in enumerate(values):
            if not self.is_value_not_empty(value):
                continue
            one_budget = self.calculate_budget_for_value(value)

            if not self.is_enough_budget(state, one_budget + messages_budget):
                is_fully_fitted = False
                logger.info(
                    f"not enough budget:{chat_value.prompt[:30]} with index {i},"
                    " for while_fits, breaking the loop"
                )
                left_budget = state.left_budget - messages_budget
                if (
                    chat_value.continue_if_doesnt_fit
                    and left_budget > settings.EXPECTED_MIN_BUDGET_FOR_VALUABLE_INPUT
                ):
                    continue
                break
            messages_budget += one_budget
            values_to_add.append(value)
            if value.ref_name and value.ref_value:
                state.references[value.ref_name].append(value.ref_value)
        if is_fully_fitted and chat_value.label:
            state.fully_fitted_pipitas.add(chat_value.label)
        if add_in_reverse_order:
            values_to_add = values_to_add[::-1]
        return messages_budget, values_to_add

    def is_enough_budget(self, state: State, required_budget: int) -> bool:
        return state.left_budget >= required_budget

    def add_values_in_one_message(
        self,
        values: list[ChatMessage],
        chat_value: ChatsEntity,
        state: State,
    ) -> CallingMessages:
        one_message_budget = 0
        one_message = None
        is_fully_fitted = True
        if not values:
            logger.info(
                f"[{self.task_name}]: values to add is empty {chat_value.prompt}"
            )

        for i, value in enumerate(values):
            if not self.is_value_not_empty(value):
                continue
            one_budget = self.calculate_budget_for_value(value)
            if not self.is_enough_budget(state, one_budget + one_message_budget):
                is_fully_fitted = False
                logger.info(
                    f"not enough budget:\n{chat_value.prompt[:30]} with index {i},"
                    f" for while_fits, breaking the loop."
                    f" Budget required: {one_budget}, "
                    f"left: {state.left_budget - one_message_budget}"
                )

                left_budget = state.left_budget - one_message_budget
                if (
                    chat_value.continue_if_doesnt_fit
                    and left_budget > settings.EXPECTED_MIN_BUDGET_FOR_VALUABLE_INPUT
                ):
                    continue
                break

            one_message_budget += one_budget
            if one_message:
                one_message.content += "\n" + value.content
            else:
                one_message = value
            if value.ref_name and value.ref_value:
                state.references[value.ref_name].append(value.ref_value)
        if is_fully_fitted and chat_value.label:
            state.fully_fitted.append(chat_value.label)
        return one_message_budget, [] if not one_message else [one_message]

    @property
    def left_budget(self) -> int:
        return self.model_max_tokens - self.min_sample_tokens - self.safe_gap_tokens

    def calculate_budget_for_value(self, value: ChatMessage) -> int:
        content = len(self.encoding.encode(value.content))
        role = len(self.encoding.encode(value.role))
        tool_calls = len(self.encoding.encode(value.tool_calls.get("name", "")))
        arguments = len(self.encoding.encode(value.tool_calls.get("arguments", "")))
        return content + role + tool_calls + arguments + settings.SAFE_GAP_PER_MSG

    def is_value_not_empty(self, value: ChatMessage) -> bool:
        if not value:
            return False
        if value.content is None:
            return False
        return True

    def add_values(
        self,
        values: t.List[ChatMessage],
        state: State,
    ) -> t.Tuple[int, t.List[ChatMessage]]:
        budget = 0
        result = []

        for value in values:
            if not self.is_value_not_empty(value):
                logger.info(f"[{self.task_name}]: is_value_not_empty failed {value}")
                continue
            budget += self.calculate_budget_for_value(value)
            result.append(value)
            if value.ref_name and value.ref_value:
                state.references[value.ref_name].append(value.ref_value)
        return budget, result

    def __str__(self) -> str:
        result = ""
        for chat_value in self.pipe:
            result += f"{chat_value}\n"
        return result

    def to_dict(self) -> dict:
        return [chat_value.to_dict() for chat_value in self.pipe]



from flow_prompt.prompt.flow_prompt import FlowPrompt
from flow_prompt.ai_models import behaviour
from flow_prompt.secrets import AZURE_KEYS, OPENAI_API_KEY, OPENAI_ORG
from flow_prompt.ai_models.openai.behaviours import (
    OPENAI_GPT4_0125_PREVIEW_BEHAVIOUR,
    OPENAI_GPT4_1106_PREVIEW_BEHAVIOUR,
    OPENAI_GPT4_1106_VISION_PREVIEW_BEHAVIOUR,
    OPENAI_GPT4_BEHAVIOUR,
    OPENAI_GPT4_32K_BEHAVIOUR,
    OPENAI_GPT3_5_TURBO_0125_BEHAVIOUR,
    OPENAI_GPT3_5_TURBO_INSTRUCT_BEHAVIOUR,
)



import logging
import typing as t
from collections import defaultdict
from dataclasses import dataclass, field

from flow_prompt.prompt.chat import ChatCondition, ChatsEntity

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class BasePrompt:
    priorities: t.Dict[int, t.List[ChatsEntity]] = field(
        default_factory=lambda: defaultdict(list)
    )
    pipe: t.List[str] = field(default_factory=list)
    functions: t.List[dict] = None

    def add(
        self,
        content: str = "",
        role: str = "user",
        name: t.Optional[str] = None,
        tool_calls: t.Dict[str, str] = None,
        priority: int = 0,
        required: bool = False,
        if_exists: t.Optional[str] = None,
        if_not_exist: t.Optional[str] = None,
        is_multiple: bool = False,
        while_fits: bool = False,
        add_in_reverse_order: bool = False,
        in_one_message: bool = False,
        continue_if_doesnt_fit: bool = False,
        add_if_fitted_labels: t.List[str] = None,
        label: t.Optional[str] = None,
        presentation: t.Optional[str] = None,
        last_words: t.Optional[str] = None,
    ):
        condition = ChatCondition(if_exists, if_not_exist)
        chat_value = ChatsEntity(
            role=role,
            content=(content or ""),
            name=name,
            tool_calls=tool_calls,
            priority=priority,
            required=required,
            condition=condition,
            is_multiple=is_multiple,
            while_fits=while_fits,
            add_in_reverse_order=add_in_reverse_order,
            in_one_message=in_one_message,
            continue_if_doesnt_fit=continue_if_doesnt_fit,
            add_if_fitted_labels=add_if_fitted_labels,
            label=label,
            presentation=presentation,
            last_words=last_words,
        )
        self.priorities[priority].append(chat_value)
        self.pipe.append(chat_value._uuid)

    def add_function(self, function: dict):
        if not self.functions:
            self.functions = []
        self.functions.append(function)



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEMP_SCRIPTS_DIR = os.environ.get(
    "FLOW_PROMPT_TEMP_SCRIPTS_DIR", os.path.join(BASE_DIR, "temp_scripts")
)
SAVE_PROMPTS_LOCALLY = os.environ.get("FLOW_PROMPT_SAVE_PROMPTS_LOCALLY", False)
ENVIRONMENT = os.environ.get("FLOW_PROMPT_ENVIRONMENT", "prod")

DEFAULT_MAX_BUDGET = os.environ.get("FLOW_PROMPT_DEFAULT_MAX_BUDGET", 16000)
DEFAULT_SAMPLE_MIN_BUDGET = os.environ.get("FLOW_PROMPT_DEFAULT_ANSWER_BUDGET", 3000)
DEFAULT_PROMPT_BUDGET = os.environ.get(
    "FLOW_PROMPT_DEFAULT_PROMPT_BUDGET", DEFAULT_MAX_BUDGET - DEFAULT_SAMPLE_MIN_BUDGET
)

EXPECTED_MIN_BUDGET_FOR_VALUABLE_INPUT = os.environ.get(
    "FLOW_PROMPT_EXPECTED_MIN_BUDGET_FOR_VALUABLE_INPUT", 100
)

SAFE_GAP_TOKENS: int = os.environ.get("FLOW_PROMPT_SAFE_GAP_TOKENS", 100)
SAFE_GAP_PER_MSG: int = os.environ.get("FLOW_PROMPT_SAFE_GAP_PER_MSG", 4)
DEFAULT_ENCODING = "cl100k_base"

USE_API_SERVICE = os.environ.get("FLOW_PROMPT_USE_API_SERVICE", None)

CACHE_PROMPT_FOR_EACH_SECONDS = os.environ.get(
    "FLOW_PROMPT_CACHE_PROMPT_FOR_EACH_SECONDS", 5 * 60
)  # 5 minutes by default
PIPE_PROMPTS = {}
AI_CLIENTS = {}



import typing as t
from dataclasses import dataclass
from enum import Enum

from _decimal import Decimal

from flow_prompt.responses import AIResponse


class AI_MODELS_PROVIDER(Enum):
    OPENAI = "openai"
    AZURE = "azure"


@dataclass(kw_only=True)
class AIModel:
    max_tokens: int
    tiktoken_encoding: t.Optional[str] = "cl100k_base"
    provider: AI_MODELS_PROVIDER = None
    support_functions: bool = False

    @property
    def name(self) -> str:
        return "undefined_aimodel"

    @property
    def price_per_prompt_1k_tokens(self) -> Decimal:
        raise NotImplementedError

    @property
    def price_per_sample_1k_tokens(self) -> Decimal:
        raise NotImplementedError

    def get_params(self) -> t.Dict[str, t.Any]:
        return {}

    def call(self, *args, **kwargs) -> AIResponse:
        raise NotImplementedError



import json
import logging
import typing as t
import uuid
from dataclasses import dataclass, field

from flow_prompt.exceptions import ValueIsNotResolvedException
from flow_prompt.utils import resolve

logger = logging.getLogger(__name__)


@dataclass
class ValuesCost:
    values: t.List[str]
    cost: int


@dataclass
class ChatCondition:
    if_exists: t.Optional[str] = None
    if_not_exist: t.Optional[str] = None

    def dump(self):
        return {
            "if_exists": self.if_exists,
            "if_not_exist": self.if_not_exist,
        }

    @classmethod
    def load(cls, data):
        return cls(
            if_exists=data.get("if_exists"),
            if_not_exist=data.get("if_not_exist"),
        )


class ChatMessage:
    role: str
    content: str
    name: t.Optional[str] = None
    tool_calls: t.Dict[str, str]
    ref_name: t.Optional[str] = None
    ref_value: t.Optional[str] = None

    def is_not_empty(self):
        return bool(self.content or self.tool_calls)

    def is_empty(self):
        return not self.is_not_empty()

    def not_tool_calls(self):
        return not (self.tool_calls)

    def __init__(self, **kwargs):
        self.role = kwargs.get("role", "user")
        self.content = kwargs["content"]
        self.name = kwargs.get("name")
        self.tool_calls = kwargs.get("tool_calls") or {}

    def to_dict(self):
        result = {
            "role": self.role,
            "content": self.content,
        }
        if self.name:
            result["name"] = self.name
        if self.tool_calls:
            result["tool_calls"] = self.tool_calls
        return result


# can be multiple value
@dataclass(kw_only=True)
class ChatsEntity:
    content: str = ""
    role: str = "user"
    name: t.Optional[str] = None
    tool_calls: t.Dict[str, str] = None
    priority: int = 0
    required: bool = False
    is_multiple: bool = False
    while_fits: bool = False
    condition: ChatCondition = field(default_factory=ChatCondition)
    add_in_reverse_order: bool = False
    in_one_message: bool = False
    continue_if_doesnt_fit: bool = False
    add_if_fitted_labels: t.List[str] = None
    label: t.Optional[str] = None
    presentation: t.Optional[str] = None
    last_words: t.Optional[str] = None
    ref_name: t.Optional[str] = None
    ref_value: t.Optional[str] = None

    def __post_init__(self):
        self._uuid = uuid.uuid4().hex

    def resolve(self, context: t.Dict[str, t.Any]) -> t.List[ChatMessage]:
        result = []
        content = self.content
        if self.is_multiple:
            # should be just one value like {messages} in prompt
            prompt_value = content.strip().replace("{", "").replace("}", "").strip()
            values = context.get(prompt_value, [])
            if not values:
                return []
            if not isinstance(values, list):
                raise ValueIsNotResolvedException(
                    f"Invalid value {values } for prompt {content}. Should be multiple"
                )
            else:
                # verify that values are json list of ChatMessage
                try:
                    result = [
                        ChatMessage(**({"content": c} if isinstance(c, str) else c))
                        for c in values
                    ]
                except TypeError as e:
                    raise ValueIsNotResolvedException(
                        f"Invalid value { values } for prompt {content}. Error: {e}"
                    )
            return result

        content = resolve(content, context)
        return [
            ChatMessage(
                name=self.name,
                role=self.role,
                content=content,
                tool_calls=self.tool_calls,
                ref_name=self.ref_name,
                ref_value=self.ref_value,
            )
        ]

    def validate(self, values: t.List[ChatMessage], context: t.Dict[str, str]):
        if self.condition.if_exists and not context.get(self.condition.if_exists):
            raise ValueIsNotResolvedException(
                f"If exists condition is not met for {self.prompt}"
            )
        if self.condition.if_not_exist and context.get(self.condition.if_not_exist):
            raise ValueIsNotResolvedException(
                f"If not exists condition is not met for {self.prompt}"
            )

    def get_values(self, context: t.Dict[str, str]) -> t.List[ChatMessage]:
        try:
            values = self.resolve(context)
            self.validate(values, context)
        except Exception as e:
            logger.error(
                f"Error resolving prompt {self.prompt}, error: {e}", exc_info=True
            )
            return []
        return values

    def dump(self):
        return {
            "content": self.content,
            "role": self.role,
            "name": self.name,
            "tool_calls": self.tool_calls,
            "priority": self.priority,
            "required": self.required,
            "is_multiple": self.is_multiple,
            "while_fits": self.while_fits,
            "condition": self.condition.dump(),
            "add_in_reverse_order": self.add_in_reverse_order,
            "in_one_message": self.in_one_message,
            "continue_if_doesnt_fit": self.continue_if_doesnt_fit,
            "add_if_fitted_labels": self.add_if_fitted_labels,
            "label": self.label,
            "presentation": self.presentation,
            "last_words": self.last_words,
            "ref_name": self.ref_name,
            "ref_value": self.ref_value,
        }

    @classmethod
    def load(cls, data):
        return cls(
            content=data.get("content"),
            role=data.get("role"),
            name=data.get("name"),
            tool_calls=data.get("tool_calls"),
            priority=data.get("priority"),
            required=data.get("required"),
            is_multiple=data.get("is_multiple"),
            while_fits=data.get("while_fits"),
            condition=ChatCondition.load(data.get("condition")),
            add_in_reverse_order=data.get("add_in_reverse_order"),
            in_one_message=data.get("in_one_message"),
            continue_if_doesnt_fit=data.get("continue_if_doesnt_fit"),
            add_if_fitted_labels=data.get("add_if_fitted_labels"),
            label=data.get("label"),
            presentation=data.get("presentation"),
            last_words=data.get("last_words"),
            ref_name=data.get("ref_name"),
            ref_value=data.get("ref_value"),
        )



import logging
import random
import typing as t
from copy import copy
from dataclasses import dataclass
from time import time

from flow_prompt.ai_models.attempt_to_call import AttemptToCall
from flow_prompt.exceptions import BehaviourIsNotDefined

logger = logging.getLogger(__name__)


@dataclass
class AIModelsBehaviour:
    # if you have mutiple AI Models, you can distribute the load across them.
    # If you wish to use as a fallback attempt a model which is not in the list, you can use fallback_attempt
    attempts: list[AttemptToCall]
    fallback_attempt: AttemptToCall = None


@dataclass
class PromptAttempts:
    ai_models_behaviour: AIModelsBehaviour
    count_of_retries: t.Optional[int] = None
    count: int = 0
    current_attempt: AttemptToCall = None

    def __post_init__(self):
        if self.count_of_retries is None:
            self.count_of_retries = len(self.ai_models_behaviour.attempts) + int(
                bool(self.ai_models_behaviour.fallback_attempt)
            )

    def initialize_attempt(self, flag_increase_count: bool = True):
        if self.count > self.count_of_retries:
            raise BehaviourIsNotDefined(
                f"Count of retries {self.count_of_retries} exceeded {self.count}"
            )
        if (
            self.count == self.count_of_retries
            and self.ai_models_behaviour.fallback_attempt
        ):
            self.current_attempt = self.ai_models_behaviour.fallback_attempt
            return self.current_attempt
        sum_weight = sum(
            [attempt.weight for attempt in self.ai_models_behaviour.attempts]
        )
        random_weight = random.randint(0, sum_weight)
        for attempt in self.ai_models_behaviour.attempts:
            random_weight -= attempt.weight
            if random_weight <= 0:
                if flag_increase_count:
                    self.count += 1
                attempt.attempt_number = self.count
                self.current_attempt = copy(attempt)
                return self.current_attempt
        raise BehaviourIsNotDefined(
            f"Count of retries {self.count_of_retries} exceeded {self.count}"
        )

    def __str__(self) -> str:
        return f"Current attempt {self.current_attempt} from {len(self.attempts)}"



import logging
import typing as t
from dataclasses import dataclass

from openai import AzureOpenAI

from flow_prompt import settings
from flow_prompt.ai_models.ai_model import AI_MODELS_PROVIDER
from flow_prompt.ai_models.openai.openai_models import FamilyModel, OpenAIModel

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class AzureAIModel(OpenAIModel):
    realm: t.Optional[str]
    deployment_name: t.Optional[str]
    provider: AI_MODELS_PROVIDER = AI_MODELS_PROVIDER.AZURE
    model: t.Optional[str] = None

    def __str__(self) -> str:
        return f"{self.realm}-{self.deployment_name}-{self.family}"

    def _define_family(self):
        if self.deployment_name.startswith("davinci"):
            self.family = FamilyModel.instruct_gpt.value
        elif self.deployment_name.startswith(("gpt3", "gpt-3")):
            self.family = FamilyModel.chat.value
        elif self.deployment_name.startswith(("gpt4", "gpt-4", "gpt")):
            self.family = FamilyModel.gpt4.value
        else:
            logger.warning(
                f"Unknown family for {self.deployment_name}. Please add it obviously. Setting as GPT4"
            )
            self.family = FamilyModel.gpt4.value

    def _define_tiktoken_encoding(self):
        if self.family in (FamilyModel.chat.value, FamilyModel.gpt4.value):
            self.tiktoken_encoding = "cl100k_base"
        elif self.family == FamilyModel.instruct_gpt.value:
            self.tiktoken_encoding = ""
        else:
            logger.warning(
                f"Unknown realm for {self.deployment_name}. Please add it obviously. Setting as cl100k_base"
            )
            self.tiktoken_encoding = "cl100k_base"

    def __post_init__(self):
        if not self.family:
            if self.deployment_name.startswith("davinci"):
                self.family = FamilyModel.instruct_gpt.value
            elif self.deployment_name.startswith(("gpt3", "gpt-3")):
                self.family = FamilyModel.chat.value
            elif self.deployment_name.startswith(("gpt4", "gpt-4", "gpt")):
                self.family = FamilyModel.gpt4.value
            else:
                logger.warning(
                    f"Unknown family for {self.deployment_name}. Please add it obviously. Setting as GPT4"
                )
                self.family = FamilyModel.gpt4.value
        self.verify_client_has_creds()
        logger.info(f"Initialized AzureAIModel: {self}")

    def verify_client_has_creds(self):
        if self.realm not in settings.AI_CLIENTS[self.provider]:
            raise Exception(f"Realm {self.realm} not found in AI_CLIENTS")

    @property
    def name(self) -> str:
        return f"{self.deployment_name}-{self.realm}"

    def get_params(self) -> t.Dict[str, t.Any]:
        return {
            "model": self.deployment_name,
        }

    def get_client(self):
        return settings.AI_CLIENTS[self.provider][self.realm]



import logging
import typing as t
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum

from flow_prompt import settings
from flow_prompt.ai_models.ai_model import AI_MODELS_PROVIDER, AIModel
from flow_prompt.ai_models.openai.responses import OpenAIResponse

from .utils import raise_openai_exception

C_4K = 4096
C_8K = 8192
C_128K = 127_000
C_16K = 16384
C_32K = 32768
M_DAVINCI = "davinci"

logger = logging.getLogger(__name__)


class FamilyModel(Enum):
    chat = "GPT-3.5"
    gpt4 = "GPT-4"
    instruct_gpt = "InstructGPT"


OPEN_AI_PRICING = {
    FamilyModel.chat.value: {
        C_4K: {
            "price_per_prompt_1k_tokens": Decimal(0.0015),
            "price_per_sample_1k_tokens": Decimal(0.0020),
        },
        C_16K: {
            "price_per_prompt_1k_tokens": Decimal(0.0010),
            "price_per_sample_1k_tokens": Decimal(0.0020),
        },
        C_128K: {
            "price_per_prompt_1k_tokens": Decimal(0.01),
            "price_per_sample_1k_tokens": Decimal(0.03),
        },
    },
    FamilyModel.gpt4.value: {
        C_4K: {
            "price_per_prompt_1k_tokens": Decimal(0.03),
            "price_per_sample_1k_tokens": Decimal(0.06),
        },
        C_32K: {
            "price_per_prompt_1k_tokens": Decimal(0.06),
            "price_per_sample_1k_tokens": Decimal(0.12),
        },
        C_128K: {
            "price_per_prompt_1k_tokens": Decimal(0.01),
            "price_per_sample_1k_tokens": Decimal(0.03),
        },
    },
    FamilyModel.instruct_gpt.value: {
        M_DAVINCI: {
            "price_per_prompt_1k_tokens": Decimal(0.0015),
            "price_per_sample_1k_tokens": Decimal(0.002),
        },
    },
}


@dataclass(kw_only=True)
class OpenAIModel(AIModel):
    model: t.Optional[str]
    support_functions: bool = False
    provider: AI_MODELS_PROVIDER = AI_MODELS_PROVIDER.OPENAI
    family: str = None
    should_verify_client_has_creds: bool = True


    def __str__(self) -> str:
        return f"openai-{self.model}-{self.family}"

    def __post_init__(self):
        if self.model.startswith("davinci"):
            self.family = FamilyModel.instruct_gpt.value
        elif self.model.startswith("gpt-3"):
            self.family = FamilyModel.chat.value
        elif self.model.startswith(("gpt-4", "gpt")):
            self.family = FamilyModel.gpt4.value
        else:
            logger.warning(
                f"Unknown family for {self.model}. Please add it obviously. Setting as GPT4"
            )
            self.family = FamilyModel.gpt4.value
        if self.should_verify_client_has_creds:
            self.verify_client_has_creds()
        logger.info(f"Initialized OpenAIModel: {self}")

    def verify_client_has_creds(self):
        if self.provider not in settings.AI_CLIENTS:
            raise Exception(f"Provider {self.provider} not found in AI_CLIENTS")

    @property
    def name(self) -> str:
        return self.model

    @property
    def price_per_prompt_1k_tokens(self) -> Decimal:
        return OPEN_AI_PRICING[self.family][self.max_tokens][
            "price_per_prompt_1k_tokens"
        ]

    @property
    def price_per_sample_1k_tokens(self) -> Decimal:
        return OPEN_AI_PRICING[self.family][self.max_tokens][
            "price_per_sample_1k_tokens"
        ]

    def get_params(self) -> t.Dict[str, t.Any]:
        return {
            "model": self.model,
        }

    def call(self, messages, max_tokens, **kwargs) -> OpenAIResponse:
        if self.family in [FamilyModel.chat.value, FamilyModel.gpt4.value]:
            return self.call_chat_completion(messages, max_tokens, **kwargs)
        raise NotImplementedError(f"Openai family {self.family} is not implemented")

    def get_client(self):
        return settings.AI_CLIENTS[self.provider]

    def call_chat_completion(
        self,
        messages: t.List[t.Dict[str, str]],
        max_tokens: t.Optional[int],
        functions: t.List[t.Dict[str, str]] = [],
        **kwargs,
    ) -> OpenAIResponse:
        max_tokens = min(max_tokens, self.max_tokens)
        common_args = {
            "top_p": 1,
            "temperature": 0,
            "max_tokens": max_tokens,
            "stream": False,
        }
        kwargs = {
            **{
                "messages": messages,
            },
            **common_args,
            **self.get_params(),
            **kwargs,
        }
        if functions:
            kwargs["tools"] = functions
        try:
            client = self.get_client()
            result = client.chat.completions.create(
                **kwargs,
            )
            logger.debug(f"Result: {result.choices[0]}")
            return OpenAIResponse(
                finish_reason=result.choices[0].finish_reason,
                message=result.choices[0].message,
                content=result.choices[0].message.content,
                original_result=result,
            )
        except Exception as e:
            logger.exception("[OPENAI] failed to handle chat stream", exc_info=e)
            raise_openai_exception(e)



import logging
import typing as t
from dataclasses import dataclass
from urllib import request

from flow_prompt import settings
from flow_prompt.exceptions import NotFoundPromptException
from flow_prompt.utils import current_timestamp_ms

logger = logging.getLogger(__name__)


@dataclass
class FlowPromptServiceResponse:
    prompt_id: str = None
    actual_prompt: dict = None
    actual_prompt_hash: str = None
    prompt_is_actual: bool = False


class FlowPromptService:
    url: str = "https://flow-prompt.com/api/v1/"
    cached_prompts = {}

    def get_actual_prompt(
        self,
        api_token: str,
        prompt_id: str,
        prompt_data: dict = None,
        version: str = None,
    ) -> FlowPromptServiceResponse:
        """
        Load prompt from flow-prompt
        if the user has keys:  lib -> service: get_actual_prompt(local_prompt) -> Service:
        generates hash of the prompt;
        check in Redis if that record is the latest; if yes -> return 200, else
        checks if that record exists with that hash;
        if record exists and it's not the last - then we load the latest published prompt; - > return  200 + the last record
        add a new record in storage, and adding that it's the latest published prompt; -> return 200
        update redis with latest record;
        """
        cached_prompt = self.get_cached_prompt(prompt_id)
        if cached_prompt:
            return FlowPromptServiceResponse(
                prompt_id=prompt_id,
                actual_prompt=cached_prompt,
                prompt_is_actual=True,
            )
        url = f"{self.url}prompts/{prompt_id}/"
        if version:
            url += f"?version={version}"
        headers = {"Authorization": f"Token {api_token}"}
        response = request.post(url, headers=headers, data={"prompt": prompt_data})
        if response.status_code == 200:
            response_data = response.json()
            prompt_data = response_data.get("prompt", prompt_data)
            # update cache
            self.cached_prompts[prompt_id] = {
                "prompt": response,
                "timestamp": current_timestamp_ms(),
            }
            # returns 200 and the latest published prompt, if the local prompt is the latest, doesn't return the prompt
            return FlowPromptServiceResponse(
                prompt_id=prompt_id,
                actual_prompt=response,
                actual_prompt_hash=response_data["actual_prompt_hash"],
                prompt_is_actual=response_data["prompt_is_actual"],
            )
        else:
            raise NotFoundPromptException(response.json()["detail"])

    def get_cached_prompt(self, prompt_id: str) -> dict:
        cached_data = self.cached_prompts.get(prompt_id)
        if not cached_data:
            return None
        cached_delay = current_timestamp_ms() - cached_data.get("timestamp")
        if cached_delay < settings.CACHE_PROMPT_FOR_EACH_SECONDS * 1000:
            return cached_data.get("prompt")
        return None

    def save_user_interaction(
        self,
        api_token: str,
        context: dict[str, t.Any],
        prompt_data: dict[str, t.Any],
        response: dict[str, t.Any],
        metrics: dict[str, t.Any] = {},
    ):
        url = f"{self.url}user_interactions/"
        headers = {"Authorization": f"Token {api_token}"}
        data = {
            "context": context,
            "prompt_data": prompt_data,
            "response": response,
            "metrics": metrics,
        }
        response = request.post(url, headers=headers, data=data)
        if response.status_code == 200:
            return response.json()
        else:
            raise logger.error(response.json()["detail"])



import logging
import typing as t
from dataclasses import dataclass
from decimal import Decimal

from openai import AzureOpenAI, OpenAI

from flow_prompt import secrets, settings
from flow_prompt.ai_models.ai_model import AI_MODELS_PROVIDER
from flow_prompt.ai_models.attempt_to_call import AttemptToCall
from flow_prompt.ai_models.behaviour import AIModelsBehaviour, PromptAttempts
from flow_prompt.exceptions import RetryableCustomException
from flow_prompt.prompt.pipe_prompt import PipePrompt
from flow_prompt.prompt.user_prompt import UserPrompt
from flow_prompt.responses import AIResponse
from flow_prompt.services.flow_prompt import FlowPromptService
from flow_prompt.utils import current_timestamp_ms

logger = logging.getLogger(__name__)


@dataclass
class FlowPrompt:
    api_token: str = secrets.API_TOKEN
    openai_api_key: str = secrets.OPENAI_API_KEY
    openai_org: str = secrets.OPENAI_ORG
    azure_keys: t.Dict[str, str] = secrets.AZURE_KEYS

    def __post_init__(self):
        if self.openai_api_key:
            openai_client = OpenAI(
                organization=self.openai_org,
                api_key=self.openai_api_key,
            )
            logger.info(f"Initialized OpenAI client: {openai_client}")
            settings.AI_CLIENTS[AI_MODELS_PROVIDER.OPENAI] = openai_client
        if self.azure_keys:
            settings.AI_CLIENTS[AI_MODELS_PROVIDER.AZURE] = {}
            for realm, key_data in self.azure_keys.items():
                if realm in settings.AI_CLIENTS[AI_MODELS_PROVIDER.AZURE]:
                    raise Exception(f"Realm {realm} already exists")
                settings.AI_CLIENTS[AI_MODELS_PROVIDER.AZURE][realm] = AzureOpenAI(
                    api_version=key_data.get("api_version", "2023-07-01-preview"),
                    azure_endpoint=key_data["url"],
                    api_key=key_data["key"],
                )
                logger.info(f"Initialized Azure client for {realm} {key_data['url']}")

    def call(
        self,
        prompt_id: str,
        context: t.Dict[str, str],
        behaviour: AIModelsBehaviour,
        params: t.Dict[str, t.Any] = {},
        version: str = None,
        count_of_retries: int = None,
    ) -> AIResponse:
        """
        Call flow prompt with context and behaviour
        """
        start_time = current_timestamp_ms()
        total_price = Decimal(0)
        pipe_prompt = self.get_pipe_prompt(prompt_id, version)
        prompt_attempts = PromptAttempts(behaviour, count_of_retries=count_of_retries)

        while prompt_attempts.initialize_attempt():
            current_attempt = prompt_attempts.current_attempt
            logger.info(
                f"Calling {prompt_id}. Attempt {prompt_attempts.current_attempt}"
            )
            user_prompt = pipe_prompt.create_prompt(current_attempt)
            calling_messages = user_prompt.resolve(context)
            try:
                result = current_attempt.ai_model.call(
                    calling_messages.get_messages(),
                    calling_messages.max_sample_budget,
                    **params,
                )
                price_of_call = self.get_price(
                    current_attempt,
                    self.calculate_budget_for_text(
                        user_prompt, result.get_message_str()
                    ),
                    calling_messages.prompt_budget,
                )
                self.save_user_interaction(
                    pipe_prompt.dump(),
                    context,
                    result,
                    {
                        "attempt_number": current_attempt.attempt_number,
                        "price": price_of_call,
                        "latency": current_timestamp_ms() - start_time,
                    },
                )
                return result
            except RetryableCustomException as e:
                logger.warning(f"Retryable error: {e}")

    def get_pipe_prompt(self, prompt_id: str, version: str = None) -> PipePrompt:
        """
        if the user has keys:  lib -> service: get_actual_prompt(local_prompt) -> Service:
        generates hash of the prompt;
        check in Redis if that record is the latest; if yes -> return 200, else
        checks if that record exists with that hash;
        if record exists and it's not the last - then we load the latest published prompt; - > return  200 + the last record
        add a new record in storage, and adding that it's the latest published prompt; -> return 200
        update redis with latest record;
        """
        if settings.USE_API_SERVICE is not None:
            prompt_data = None
            prompt = settings.PIPE_PROMPTS.get(prompt_id)
            if prompt is None:
                prompt_data = prompt.dump()
            response = FlowPromptService.get_actual_prompt(
                self.api_token, prompt_id, prompt_data, version
            )
            if response.prompt_is_actual:
                return prompt
            else:
                return PipePrompt.load(response.actual_prompt)
        else:
            return settings.PIPE_PROMPTS[prompt_id]

    def save_user_interaction(
        self,
        prompt_data: dict,
        context: dict[str, t.Any],
        result: AIResponse,
        metrics: dict[str, t.Any] = {},
    ) -> None:
        """
        Save user interaction to flow-prompt
        """
        if settings.USE_API_SERVICE is not None:
            FlowPromptService.save_user_interaction(
                self.api_token, prompt_data, context, result, metrics
            )

    def calculate_budget_for_text(self, user_prompt: UserPrompt, text: str) -> int:
        if not text:
            return 0
        return len(user_prompt.encoding.encode(text))

    def _decimal(self, value) -> Decimal:
        return Decimal(value).quantize(Decimal(".00001"))

    def get_price(
        self, attempt: AttemptToCall, sample_budget: int, prompt_budget: int
    ) -> Decimal:
        return self._decimal(
            prompt_budget * attempt.ai_model.price_per_prompt_1k_tokens / 1000
        ) + self._decimal(
            sample_budget * attempt.ai_model.price_per_sample_1k_tokens / 1000
        )



import logging
from urllib import request

import openai

from .exceptions import (
    OpenAIAuthenticationError,
    OpenAIBadRequestException,
    OpenAIChunkedEncodingError,
    OpenAIInternalException,
    OpenAIInvalidRequestError,
    OpenAiPermissionDeniedError,
    OpenAiRateLimitException,
    OpenAIResponseWasFilteredError,
    OpenAITimeoutException,
    OpenAIUnknownException,
)

logger = logging.getLogger(__name__)


def raise_openai_exception(
    exc: Exception,
) -> None:
    if isinstance(exc, request.exceptions.ChunkedEncodingError):
        raise OpenAIChunkedEncodingError()

    if isinstance(exc, openai.APITimeoutError):
        raise OpenAITimeoutException()

    if isinstance(exc, openai.BadRequestError):
        if "response was filtered" in str(exc):
            raise OpenAIResponseWasFilteredError()
        if "Too many inputs" in str(exc):
            raise OpenAiRateLimitException()
        raise OpenAIInvalidRequestError()
    if isinstance(exc, openai.RateLimitError):
        raise OpenAiRateLimitException()

    if isinstance(exc, openai.AuthenticationError):
        raise OpenAIAuthenticationError()

    if isinstance(exc, openai.InternalServerError):
        raise OpenAIInternalException()

    if isinstance(exc, openai.PermissionDeniedError):
        raise OpenAiPermissionDeniedError()

    if isinstance(exc, openai.APIStatusError):
        raise OpenAIBadRequestException()

    logger.error(
        "Unknown OPENAI error, please add it in raise_openai_rate_limit_exception",
        exc_info=exc,
    )
    raise OpenAIUnknownException()



import logging
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class AIResponse:
    _response: str = ""
    original_result: object = None

    @property
    def response(self) -> str:
        return self._response



from flow_prompt.ai_models.attempt_to_call import AttemptToCall
from flow_prompt.ai_models.behaviour import AIModelsBehaviour
from flow_prompt.ai_models.openai.openai_models import (
    C_128K,
    C_4K,
    OpenAIModel,
    C_16K,
    C_32K,
)


OPENAI_GPT4_0125_PREVIEW_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-4-0125-preview",
                max_tokens=C_128K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)

OPENAI_GPT4_1106_PREVIEW_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-4-1106-preview",
                max_tokens=C_128K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)

OPENAI_GPT4_1106_VISION_PREVIEW_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-4-1106-vision-preview",
                max_tokens=C_128K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)

OPENAI_GPT4_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-4",
                max_tokens=C_4K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)

OPENAI_GPT4_32K_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-4-32k",
                max_tokens=C_32K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)

OPENAI_GPT3_5_TURBO_0125_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-3.5-turbo-0125",
                max_tokens=C_16K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)

OPENAI_GPT3_5_TURBO_INSTRUCT_BEHAVIOUR = AIModelsBehaviour(
    attempts=[
        AttemptToCall(
            ai_model=OpenAIModel(
                model="gpt-3.5-turbo-instruct",
                max_tokens=C_4K,
                support_functions=True,
                should_verify_client_has_creds=False,
            ),
            weight=100,
        ),
    ]
)



class FlowPromptException(Exception):
    pass


class RetryableCustomException(FlowPromptException):
    pass


class BehaviourIsNotDefined(FlowPromptException):
    pass


class ValueIsNotResolvedException(FlowPromptException):
    pass


class NotEnoughBudgetException(FlowPromptException):
    pass


class NotFoundPromptException(FlowPromptException):
    pass



import logging
from copy import deepcopy
from dataclasses import dataclass

from flow_prompt import settings
from flow_prompt.ai_models.attempt_to_call import AttemptToCall
from flow_prompt.prompt.base_prompt import BasePrompt
from flow_prompt.prompt.chat import ChatsEntity
from flow_prompt.prompt.user_prompt import UserPrompt
from flow_prompt.settings import PIPE_PROMPTS

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class PipePrompt(BasePrompt):
    """
    PipePrompt is a class that represents a pipe of chats that will be used to generate a prompt.
    You can add chats with different priorities to the pipe thinking just about the order of chats.
    When you initialize a Prompt, chats will be sorted by priority and then by order of adding.
    """

    id: str
    max_tokens: int = None
    min_sample_tokens: int = settings.DEFAULT_SAMPLE_MIN_BUDGET
    reserved_tokens_budget_for_sampling: int = None

    def __post_init__(self):
        PIPE_PROMPTS[self.id] = self

    def get_max_tokens(self, ai_attempt: AttemptToCall) -> int:
        if self.max_tokens:
            return min(self.max_tokens, ai_attempt.model_max_tokens())
        return ai_attempt.model_max_tokens()

    def create_prompt(self, ai_attempt: AttemptToCall) -> UserPrompt:
        logger.debug(
            f"Creating prompt for {ai_attempt.ai_model} with {ai_attempt.attempt_number} attempt"
            f"Encoding {ai_attempt.tiktoken_encoding()}"
        )
        return UserPrompt(
            pipe=deepcopy(self.pipe),
            priorities=deepcopy(self.priorities),
            tiktoken_encoding=ai_attempt.tiktoken_encoding(),
            model_max_tokens=self.get_max_tokens(ai_attempt),
            min_sample_tokens=self.min_sample_tokens,
            reserved_tokens_budget_for_sampling=self.reserved_tokens_budget_for_sampling,
        )

    def dump(self) -> dict:
        return {
            "id": self.id,
            "max_tokens": self.max_tokens,
            "min_sample_tokens": self.min_sample_tokens,
            "reserved_tokens_budget_for_sampling": self.reserved_tokens_budget_for_sampling,
            "priorities": {
                priority: [chats_value.dump() for chats_value in chats_values]
                for priority, chats_values in self.priorities.items()
            },
            "pipe": self.pipe,
        }

    @classmethod
    def load(cls, data):
        priorities = {}
        for priority, chat_values in data["priorities"].items():
            priorities[int(priority)] = [
                ChatsEntity.load(chat_value) for chat_value in chat_values
            ]
        return cls(
            id=data["id"],
            max_tokens=data["max_tokens"],
            min_sample_tokens=data.get("min_sample_tokens"),
            reserved_tokens_budget_for_sampling=data.get(
                "reserved_tokens_budget_for_sampling"
            ),
            priorities=priorities,
            pipe=data["pipe"],
        )



import logging
import typing as t
from time import time

logger = logging.getLogger(__name__)


def current_timestamp_ms():
    return int(time() * 1000)


def resolve(prompt: str, context: t.Dict[str, str]) -> str:
    if not prompt or "{" not in prompt:
        return prompt
    # TODO: monitor how many values were not resolved and what values
    for key in context:
        prompt = prompt.replace(f"{{{key}}}", str(context[key]))
    return prompt



import json
import logging
import typing as t
from dataclasses import dataclass

from openai.types.chat import ChatCompletionMessage as Message
from openai.types.chat import ChatCompletionMessageToolCall as ToolCall

from flow_prompt.responses import AIResponse

FINISH_REASON_LENGTH = "length"
FINISH_REASON_ERROR = "error"
FINISH_REASON_FINISH = "stop"
FINISH_REASON_TOOL_CALLS = "tool_calls"

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class OpenAIResponse(AIResponse):
    content: str = ""
    finish_reason: str = ""
    message: Message = None
    exception: t.Optional[Exception] = None

    @property
    def response(self) -> str:
        return self.content or self.message.content

    def is_function(self) -> bool:
        return self.finish_reason == FINISH_REASON_TOOL_CALLS

    @property
    def tool_calls(self) -> t.List[ToolCall]:
        return self.message.tool_calls

    def get_function_name(self, tool_call: ToolCall) -> t.Optional[str]:
        if tool_call.type != "function":
            logger.error(f"function.type is not function: {tool_call.type}")
            return None
        return tool_call.function.name

    def get_function_args(self, tool_call: ToolCall) -> t.Dict[str, t.Any]:
        if not self.is_function() or not tool_call.function:
            return {}
        arguments = tool_call.function.arguments
        try:
            return json.loads(arguments)
        except json.JSONDecodeError as e:
            logger.info("Failed to parse function arguments", exc_info=e)
            return {}

    def is_reached_limit(self) -> bool:
        return self.finish_reason == FINISH_REASON_LENGTH

    def to_dict(self) -> t.Dict[str, str]:
        return {
            "finish_reason": self.finish_reason,
            "message": self.message,
            "content": self.content,
        }

    def get_message_str(self) -> str:
        return self.message.model_dump_json(indent=2)

    def __str__(self) -> str:
        result = (
            f"finish_reason: {self.finish_reason}\n"
            f"message: {self.get_message_str()}\n"
        )
        return result
